"""
This package manages the geographic information associated with data points in both swath and grid rasters.
"""
# noinspection PyUnusedImports
from .constants import *
from .CRS import CRS, WGS84
from .bbox import BBox
from .spatial_geometry import SpatialGeometry
from .coordinate_array import CoordinateArray
from .vector_geometry import VectorGeometry, SingleVectorGeometry, MultiVectorGeometry
from .point import Point
from .multi_point import MultiPoint
from .polygon import Polygon
from .multi_polygon import MultiPolygon
from .wrap_geometry import wrap_geometry
from .kdtree import KDTree
from .raster_geometry import RasterGeometry
from .raster_geolocation import RasterGeolocation
from .out_of_bounds_error import OutOfBoundsError
from .raster_grid import RasterGrid
from .raster import Raster
from .multi_raster import MultiRaster
from .where import where
from .clip import clip
from .center_aeqd import center_aeqd
from .local_UTM_proj4 import local_UTM_proj4
from .linear_downscale import linear_downscale
from .bias_correct import bias_correct

__author__ = "Gregory Halverson"
