def test_import_rasters():
    import rasters as rt

def test_import_raster():
    from rasters import Raster

def test_import_multi_raster():
    from rasters import MultiRaster

def test_import_spatial_geometry():
    from rasters import SpatialGeometry

def test_import_coordinate_array():
    from rasters import CoordinateArray

def test_import_vector_geometry():
    from rasters import VectorGeometry

def test_import_single_vector_geometry():
    from rasters import SingleVectorGeometry

def test_import_multi_vector_geometry():
    from rasters import MultiVectorGeometry

def test_import_point():
    from rasters import Point

def test_import_multi_point():
    from rasters import MultiPoint

def test_import_polygon():
    from rasters import Polygon

def test_import_multi_polygon():
    from rasters import MultiPolygon

def test_import_kdtree():
    from rasters import KDTree

def test_import_raster_geometry():
    from rasters import RasterGeometry

def test_import_raster_geolocation():
    from rasters import RasterGeolocation

def test_import_out_of_bounds_error():
    from rasters import OutOfBoundsError

def test_import_raster_grid():
    from rasters import RasterGrid
